Require Import ssreflect ssrbool ssrnat eqtype fintype tuple finfun ssrfun seq.
Require Import pred prelude idynamic ordtype finmap pcm coding unionmap heap.
Require Import feap state zigzag world inject rely atomic autosplit.
Require Import privstate model lemmas stsep histories getter array spcm.
Require Import Eqdep alloc treiber histories flatcombine.  

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(*******************************************)
(* Instantiating flat combiner with stacks *)
(*******************************************)

Section StackCombine.
Variable a : ptr. (* origin of the side-channel array *)
Variable n : nat. (* the size of the array *)
Variable lk : ptr. (* the lock for the shared resource *)
Variable fc : nat. (* the label for the flat-combine concurroid *)
Variable sent : ptr. (* the sentinel for the stack *)
Variable A : encoded_set. (* the type of stack elements *)

Notation st := [encoded_set of seq A].

Definition P_stack s1 s2 := exists e : A, s1 = e::s2 \/ s2 = e::s1.

Definition inv (H : hist st) (h : heap) := 
  exists p h' ls, 
  [/\ h = sent :-> p \+ h', llist p ls h', [ls in H at last_key H], 
      cont P_stack H, complete H & valid h].

Lemma prec_inv (H : hist st) : precise (inv H).
Proof.
move=>h1 h2 t1 t2 V E. case=>p1 [h1'][ls1][E1 L1 K1 C1 Cp1 V1]. 
case=>p2 [h2'][ls2][E2 L2 K2 C2 Cp2 V2].
rewrite E1 E2 -!joinA in E V *.
case/(hcancelV V): E=>? {V} V. subst p2.
move/(llist_prec V). move/(_ _ _ _ _ L1 L2).
case=>???. subst ls2 h2' t2. by [].
Qed.

Structure code := pushc | popc. 

Definition inT c : Type := if c is pushc then A else unit. 
Definition outT c : Type := if c is pushc then unit else option A.

Definition specR c := 
  if c is pushc as c' return inT c' -> outT c' -> hist st -> hist st -> _ 
  then fun e _ H g => 
     exists ls, [ls in H at last_key H] /\ g = fresh H \\-> (ls, e::ls)   
  else fun _ v H g => 
     if v is Some e then 
       exists ls, [e :: ls in H at last_key H] /\ g = fresh H \\-> (e::ls, ls)
     else [[::] in H at last_key H] /\ g = Unit.

Lemma uniqR m (v : inT m) (w : outT m) g g1' g2' : 
        valid (g \+ g1') -> valid (g \+ g2') -> 
        specR v w g g1' -> specR v w g g2' -> g1' = g2'.
Proof.
move=>_ _; case: m v w=>/= [v _|_ [w|[_ -> [_ ->] //]]];
case=>ls1 [[x1 F1 ->]][ls2][[x2 F2 ->]]; by rewrite F1 in F2; case: F2=>_ ->. 
Qed.

Variable seq_push : forall pv al asent alk (e : A), 
  {g}, STsep [entangle (Priv pv) (AllocCAS al asent alk)]
    (fun i => inv g (self (pvg pv) i), 
     fun r m => exists g', 
       [/\ valid (g \+ g'), inv (g \+ g') (self (pvg pv) m) &
            @specR pushc e r g g']).

Variable seq_pop : forall pv al asent alk (_ : unit), 
  {g}, STsep [entangle (Priv pv) (AllocCAS al asent alk)]
    (fun i => inv g (self (pvg pv) i),
     fun r m => exists g', 
       [/\ valid (g \+ g'), inv (g \+ g') (self (pvg pv) m) & 
           @specR popc tt r g g']).

Definition interp (pv al : nat) (asent alk : ptr) c := 
  if c is pushc as c' return forall v : inT c', {g}, 
     STsep [entangle (Priv pv) (AllocCAS al asent alk)]
       (fun i => inv g (self (pvg pv) i), 
        fun r m => exists g', 
          [/\ valid (g \+ g'), inv (g \+ g') (self (pvg pv) m) &
              @specR c' v r g g'])
  then @seq_push pv al asent alk else @seq_pop pv al asent alk.

Definition T := 
  @FlatCombineT a lk n fc [encoded_pcm of hist st] inv prec_inv 
                code inT outT specR interp uniqR.

Variables (pv al : nat) (asent alk : ptr).
Notation PA := (entangle (Priv pv) (AllocCAS al asent alk)).
Notation W := (entangle PA (FlatCombine T)).
Local Notation geth gc s := ((joint gc s).1).
Local Notation getg gc s := ((joint gc s).2).
Notation pv_getso s := (getso (pvg pv) s).
Notation pv_self s := (self (pvg pv) s).
Notation pv_other s := (other (pvg pv) s).
Notation pv_joint s := (joint (pvg pv) s).
Notation fc_getso s := (getso (fcg T) s).
Notation fc_getj s := (getj (fcg T) s).
Notation fc_self s := (self (fcg T) s).
Notation fc_other s := (other (fcg T) s).
Notation fc_joint s := (joint (fcg T) s).
Notation fc_geth s := (geth (fcg T) s).
Notation fc_getg s := (getg (fcg T) s).

Program Definition fc_push (tid : 'I_n) (e : A) : {h : hist st}, STsep [W] 
  (fun i => [/\ pv_self i = Unit, fc_self i = (#nat_of_ord tid, up nown, Unit),
                [h <<= getU (fc_other i) \+ sum_up (fc_getg i)] & 
                @no_reqs T tid i],
   fun (_ : unit) m => exists t ls, 
      [/\ pv_self m = Unit, 
          fc_self m = (#nat_of_ord tid, up nown, t \\-> (ls, e :: ls)),
          [h <<= getU (fc_other m) \+ sum_up (fc_getg m)], last_key h < t & 
          @no_reqs T tid m]) := 
  Do (@flat_combine' T pv al asent alk tid pushc e).
Next Obligation.
apply: gh=>i h [Hp Hf Eg N] C; apply: (gh_ex h); apply: val_do.
- by split=>//; rewrite Hf /= unitL; case: Eg=>_ [x ->]; exists x. 
case=>m [g'][g1][Hp1 Hf1] Eg1 Eg2 [ls][K] Eg1' N1 Cm.
have V1 : valid g1 by case: K=>x /find_some /dom_valid. 
have Vi : valid (getU (fc_other i) \+ sum_up (fc_getg i)). 
- case: (C)=>_ [s1][s2][?] Cs1 Cs2; subst i. 
  rewrite !(getC Cs2 C); case: Cs2=>_ _ [r][b][ha][hr][_ _ _ _ V' _].
  by move: V'; rewrite -!joinA; move/validR. 
have Vm : valid (getU (fc_self m \+ fc_other m) \+ sum_up (fc_getg m)).
- case: (Cm)=>_ [s1][s2][?] Cs1 Cs2; subst m. 
  rewrite !(getC Cs2 Cm) /= in Hf1 Eg1 *. 
  by case: Cs2=>_ _ [r][b][ha][hr][]. 
have {Eg2} Eg2 : [h <<= g1].
- split; first by case: Eg=>V' _; rewrite V' Vi V1.
  by case: Eg2=>x <-; exists x. 
have {Eg1} Eg1 : [g1 <<= g' \+ getU (fc_other m) \+ sum_up (fc_getg m)].
- split; first by rewrite V1; rewrite Hf1 /= in Vm; rewrite Vm.
  by case: Eg1=>x <-; exists x.
exists (fresh g1), ls; split=>//; last 1 first.
- by apply: histK Eg2.
- by rewrite Hf1 -Eg1'.
apply: hist_trans Eg2 _; rewrite -!joinA joinC in Eg1.
by apply: hist_plus_mono2 Eg1; rewrite Eg1' valid_fresh.  
Qed.


Program Definition fc_pop (tid : 'I_n) : {h : hist st}, STsep [W]
  (fun i => [/\ pv_self i = Unit, fc_self i = (#nat_of_ord tid, up nown, Unit),
                [h <<= getU (fc_other i) \+ sum_up (fc_getg i)] &
                @no_reqs T tid i],
   fun v m => 
     [h <<= getU (fc_other m) \+ sum_up (fc_getg m)] /\
     if v is Some e then 
       exists t ls, last_key h < t /\ 
         fc_self m = (#nat_of_ord tid, up nown, t \\-> (e :: ls, ls))
     else exists t, 
       [/\ last_key h <= t, fc_self m = (#nat_of_ord tid, up nown, Unit) & 
           [[::] in getU (fc_other m) \+ sum_up (fc_getg m) at t]]) := 
  Do (@flat_combine' T pv al asent alk tid popc tt).
Next Obligation.
apply: gh=>i h [Hp Hf Eg N] C; apply: (gh_ex h); apply: val_do.
- by split=>//; rewrite Hf /= unitL; case: Eg=>V [x ->]; exists x. 
move=>x m [g'][g1][Hp1 Hf1 Eg1 Eg2] X N1 Cm.
have V1 : valid g1 by case: x X=>[a' [ls']|][][x] /find_some /dom_valid. 
have Vi : valid (getU (fc_other i) \+ sum_up (fc_getg i)). 
- case: (C)=>_ [s1][s2][?] Cs1 Cs2; subst i. 
  rewrite !(getC Cs2 C); case: Cs2=>_ _ [r][b][ha][hr][_ _ _ _ V' _].
  by move: V'; rewrite -!joinA; move/validR. 
have Vm : valid (getU (fc_self m \+ fc_other m) \+ sum_up (fc_getg m)).
- case: (Cm)=>_ [s1][s2][?] Cs1 Cs2; subst m. 
  rewrite !(getC Cs2 Cm) /= in Hf1 Eg1 *. 
  by case: Cs2=>_ _ [r][b][ha][hr][]. 
have {Eg2} Eg2 : [h <<= g1].
- split; first by case: Eg=>V' _; rewrite V' Vi V1.
  by case: Eg2=>z <-; exists z.
have {Eg1} Eg1 : [g1 <<= g' \+ getU (fc_other m) \+ sum_up (fc_getg m)].
- split; first by rewrite V1; rewrite Hf1 /= in Vm; rewrite Vm.
  by case: Eg1=>z <-; exists z.
split.
- apply: hist_trans Eg2 _; rewrite -!joinA joinC in Eg1. 
  apply: hist_plus_mono2 Eg1; case: x X => [a' [ls']|][_] ->;
  by [rewrite valid_fresh | rewrite unitR].
case: x X=>[a' [ls']|][K] Eg'.
- by exists (fresh g1), ls'; rewrite -Eg'; split=>//; apply: histK Eg2.
exists (last_key g1); split; first by [apply: histK Eg2]; first by rewrite -Eg'.
by apply: histE K; rewrite Eg' unitL in Eg1.
Qed.

End StackCombine.



